export class UnmatchedTagError extends Error {
    constructor(tag) {
        super(`unmatched tag in union: ${tag}`)
    }
}

// TODO: verify patterns against union
/**
 * @template Type, Value
 * @constructor
 * @param {Type} type
 * @param {Iterable<Value>} values
 */
export const Tag = (type, values) => ({
    /**
     * @template T
     * @typedef {(...args: T[]) => void}) Pattern
     * @param {Object.<string, Pattern>} patterns
     * @throws {UnmatchedTagError}
     */
    fold: patterns => {
        if (type in patterns) {
            return patterns[type](...values)
        } else if ('_' in patterns) {
            return patterns._()
        } else {
            throw new UnmatchedTagError(type)
        }
    },
    toString: () => `${type}(${values.join(', ')})`
})

export const Union = types => {
    const result = {}

    const len = types.length
    for (let i = 0; i < len; i++) {
        const type = types[i]
        result[type] = (...values) => Tag(type, values)
    }

    return result
}

